# -*- tcl -*- \
    last modified <20111225.1528.02>
# This is the replacement for the ftp code that used to be in C.
#
proc FTP_CreateServerSocket { ftpI } {
  # There can be more than one ftp connection at a time...
  # but only two ports open per connection...
  global ftp
  if  { $ftp($ftpI,debug) } {
    puts "[info exists ftp($ftpI,server1)] --$ftp($ftpI,state)"
  }

  if { $ftp(server1) == "" } {
    set accept "1" 
  } elseif { $ftp(server2) == "" } {
    set accept "2" 
  } else {
    FTP_Error $ftpI "Oops, can not handle more than two open ports..."
  }
  set sockid [socket -server FTP_Accept$accept -myaddr $ftp($ftpI,local_ip) 0]
  set info [fconfigure $sockid -sockname]
  set port [lindex $info 2]
  # The first entry in the list is the master port, 
  # the rest are slave sockets f0r that port with different port numbers.
  set ftp(server$accept) $sockid
  set ftp($port) $ftpI
  if  { $ftp($ftpI,debug) } {
    puts " FTP_CreateServerSocket exits  [lindex $info 0],$port $sockid"
  }
  return "[lindex $info 0],$port $sockid"
}

proc FTP_Accept1 { sockid address port } { 
  global ftp
  # The first entry in the list is the master port, 
  # the rest are slave sockets f0r that port with different port numbers.

  
  lappend ftp(server1) $sockid
#  puts "FTP_Accept1 $sockid $address $port $ftp(server1)]"
}


proc FTP_Accept2 { sockid address port } { 
  global ftp
  # The first entry in the list is the master port, 
  # the rest are slave sockets f0r that port with different port numbers.

  lappend ftp(server1) $sockid
}

proc FTP_AcceptConnect { sockid } {
  global ftp config
  set info [fconfigure $sockid -sockname]
  set port [lindex $info 2]
  set ftpI $ftp($port)
  if  { $ftp($ftpI,debug) } {
    puts "FTP_AcceptConnect --$ftp($ftpI,state)"
  }
  if { [lrange $ftp(server1) 0 0] == $sockid } {
    set accept "1"
  } elseif { [lrange $ftp(server2) 0 0] == $sockid } {
    set accept "2"
  } else {
    FTP_Error $ftpI "structure not found for socket $sockid"
  }

  if  { $ftp($ftpI,debug) } {
    puts "ftp(server$accept) is $ftp(server$accept)"
  }
  if { [llength $ftp(server$accept)] < 2 } {
    set timerid [after [expr $config(ftp,timeout) * 1000 ] \
		     "FTP_Accept$accept {} {} {}"]
    if  { $ftp($ftpI,debug) } {
      puts "FTP_AcceptConnect --waiting on ftp(server$accept)"
    }   
    vwait ftp(server$accept)
    after cancel $timerid
  }
  set sock [lrange $ftp(server$accept) 1 1]
  if { $sock == "" } {
    FTP_Error $ftpI "Time out while waiting for server data"
  }
  set ftp(server$accept) [lreplace $ftp(server$accept) 1 end]
  fconfigure $sock -translation {binary binary}
  if  { $ftp($ftpI,debug) } {
    puts "FTP_AcceptConnect --returning $sock"
  }
  return $sock
}


proc FTP_Socket { host port  } {
  
  set sockid [socket $host $port]
  set info [fconfigure $sockid -sockname]
  fconfigure $sockid -translation {binary binary}
  return "$sockid [lindex $info 0]"
}

proc FTP_ReadText {from timeout } {
  global readresult readresultdone glob
  set readresult ""
  set readresultdone 0

  fconfigure $from -blocking 0
#  fconfigure $from -eofchar \n
  chan event $from readable "FTP_GetLine $from"
  set timer [after [expr $timeout * 1000] "set readresultdone 2"]
  vwait readresultdone
  after cancel $timer
  if { $readresultdone == 2 } {
    if  { $glob(ftp,debug) } {
      puts "FTP_ReadText: >TIME OUT <"
    }
    return -code return  "Timeout when communicating with FTP server."
  }
  if  { $glob(ftp,debug) } {
    puts "FTP_ReadText: $readresult"
  }
  return $readresult
}

proc FTP_GetLine { from } {
  global readresult readresultdone ftp glob
#  if { $ftp($ftpI,debug) != 0 } {
#    puts "in FTP_GetLine"
#  }
  while { 1 } {


    set char [read $from 1]
    # if we have a new line, toss it and set the done flag
    if { $char == "\n" || [eof $from] } {
      chan event $from readable ""
      set readresultdone 1
      if { $glob(ftp,debug) != 0 } {
	puts "FTP_Getline: $from -> $readresult"
      }
      return
    }
    # if we have a '\r' toss it
    if { $char != "\r" } {
      set readresult $readresult$char
    }
  }
}


proc FTP_WriteText { sockid output } {
  global glob
  if { $glob(ftp,debug) != 0 } {
    puts "FTP_WriteText:  $output -> $sockid"
  }
  puts -nonewline $sockid $output
  flush $sockid
}

proc FTP_OpenFile { fname  mode } {

  # For histerical reasons... mode takes the following values:
  # r      Open read only (O_RDONLY)
  # w      Open write only (O_CREAT|O_WRONLY|O_TRUNC)
  # w+     Open write only append (O_CREAT|O_WRONLY|O_APPEND)

  if { $mode == "w" } {
    return [open $fname { CREAT WRONLY TRUNC BINARY}]
  } elseif { $mode == "w+" } {
    return [open $fname { CREAT WRONLY APPEND BINARY}]
  }
  return [open $fname {RDONLY BINARY}]    
}

proc FTP_Close { sockid } {
  global ftp
  if { [lrange $ftp(server1) 0 0] == $sockid } {
    foreach sock $ftp(server1) {
#      puts "Closing server1 $sock"
      close $sock
    }
    set ftp(server1) ""
#    puts "Reset server1 $ftp(server1)"
    return
  }
  if { [lrange $ftp(server2) 0 0] == $sockid } {
    foreach sock $ftp(server2) {
#      puts "Closing server2 $sock"
      close $sock
    }
    set ftp(server2) ""
    return
  }
#  puts "Closing  $sockid 1-> $ftp(server1) 2-> $ftp(server2)"
  close $sockid
}

proc FTP_Copy { from to size timeout  } {
  global ftp_copyresult ftp_copyerror glob
  set ftp_copyresult 0
  set ftp_copyerror ""

  # It would appear the vwait only works with variables that 
  # are not lists.  We use ftp_copyresult to wake us up either
  # when the copy complets or times out, which ever.

  set timer [after [expr $timeout * 1000] \
		 {global ftp_copyresult ftpcopyerror; \
		      set ftp_copyerror \
		      "Timeout when communicating with FTP server." ;\
		      set ftp_copyresult -1}]

  chan copy $from $to -size $size -command FTP_CopyCallback
  vwait ftp_copyresult
  if { $ftp_copyerror  != "" } {
    return code return $ftpcopyerror
  }
  if { $glob(ftp,debug) } {
    puts "FTP_copy: [lindex $ftp_copyresult 0] "
  }
  return $ftp_copyresult
}

proc FTP_CopyCallback { count { errorstring ""} } {
  global ftp_copyresult ftp_copyerror glob
  if { $glob(ftp,debug) } {
    puts "FTP_CopyCallback: $count -> $errorstring"
  }
  set ftp_copyresult $count 
  set ftp_copyerror $errorstring
}

proc GetTimeFromSecs { time } {
  global dateformat
  # Convert time in sec (UTC) to either yymmdd hh:mm:ss or ddmmyy hh:mm:ss
  return [clock format $time -format $dateformat]
}
proc GetTimeFromSecsSetFormat { format } {
  global dateformat
  # 0 -> yymmdd
  # 1 -> ddmmyy
  if { $format == 0 } {
    set dateformat "%y%m%d %H:%M:%S"
  } elseif { $format == 1 } {
    set dateformat "%d%m%y %H:%M:%S"
  } else {
    set dateformat $format
  }
}

proc GetStringFromMode { mode } {
  global glob
  if { $glob(os) == "WIN32" } {
    return [format "0x%x" $mode]
  }
#define S_ISUID  0004000
#define S_ISGID  0002000
#define S_ISVTX  0001000
#define S_IRUSR 00400
#define S_IWUSR 00200
#define S_IXUSR 00100
#define S_IRGRP 00040
#define S_IWGRP 00020
#define S_IXGRP 00010
#define S_IROTH 00004
#define S_IWOTH 00002
#define S_IXOTH 00001
  return [format "%s%s%s%s%s%s%s%s%s" \
	      [expr {$mode &  00400 ? "r" : "-"}] \
	      [expr {$mode &  00200 ? "w" : "-"}] \
	      [expr {$mode & 0004000 ? "S" : \
			 [expr {$mode & 00100 ? "x" : "-"}]}\
		  ] \
	      [expr {$mode & 00040 ? "r" : "-"}] \
	      [expr {$mode & 00020 ? "w" : "-"}] \
	      [expr {$mode & 0002000 ? "S" : \
			 [expr {$mode & 00010 ? "x" : "-"}]}\
		  ] \
	      [expr {$mode &  00004 ? "r" : "-"}] \
	      [expr {$mode & 00002 ? "w" : "-"}] \
	      [expr {$mode & 0001000 ? "T" : \
			 [expr {$mode & 00001 ? "x" : "-"}]}\
		  ]\
	  ]
#  return [format "%s%s%s%s%s%s%s%s%s" \
# 	      [expr $mode & $S_IRUSR ? "{r}" : "{-}"] \
# 	      [expr $mode & $S_IWUSR ? "{w}" : "{-}"] \
# 	      [expr $mode & $S_ISUID ? "{S}" : \
# 		   [expr $mode & $S_IXUSR ? "{x}" : "{-}"]] \
# 	      [expr $mode & $S_IRGRP ? "{r}" : "{-}"] \
# 	      [expr $mode & $S_IWGRP ? "{w}" : "{-}"] \
# 	      [expr $mode & $S_ISGID ? "{S}" : \
# 		   [expr $mode & $S_IXGRP ? "{x}" : "{-}"]] \
# 	      [expr $mode & $S_IROTH ? "{r}" : "{-}"] \
# 	      [expr $mode & $S_IWOTH ? "{w}" : "{-}"] \
# 	      [expr $mode & $S_ISVTX ? "{T}" : \
# 		   [expr $mode & $S_IXOTH ? "{x}" : "{-}"]]]
}

proc ClockMilliSeconds { } {
  return [clock milliseconds]
}

proc GetEuid { } {
  global glob
  
  if { $glob(os) == "WIN32" } {
    return "1"
  }
  return "whodat"
}

proc GetUidGidString { file uid gid } {
  global glob users groups
  
  if { $glob(os) == "WIN32" } {
    return ""
  }
  if { ![info exists users($uid)] || ![info exists groups($gid)] } {
    # Need to work for it...
    if { [catch {file attributes $file} gu] != 0 } {
      # rare but happens with dead links...
      return ""
    }
    set gu [file attributes $file]
    set users($uid) [lindex $gu 3]
    set groups($gid) [lindex $gu 1]
  }
  return "$users($uid)/$groups($gid)" 
}

proc GetDF { dir } {
  # Makes a stab at getting the full size of the disk
  global glob
  set r [catch "exec df {$dir}" out]
  set oout [split $out]
  set indx [lsearch -regexp $oout {[0-9]%}]
  if { $indx != -1 } {
    incr indx -1
    while {[set free [lrange $oout $indx $indx]] == "{}" } {
      incr indx -1
    }
#    puts "$oout"
#    puts "$free"
    set indx 1

  } elseif {$glob(os) == "WIN32" &&
	    [catch {exec cmd /C dir} out] == 0 } {
    set oout [split $out "\n"]
    set free [split [lrange $oout end end]]
    set indx [lsearch -exact $free bytes]
    incr indx -1
    set free [lrange $free $indx $indx]
    set indx 0
  } else {
    LogSilent $out
    return "??"
  }
  regsub -all {[^0-9]} $free "" free
  # should now be a pure number.  If nil, use zero...
  # better than failing in the following code...
  if { $free == "" } {
    set free 0
  }
  while { $free > 1024. } {
    incr indx
    set free [expr {$free / 1024.}]
  }
#  puts "$indx $free"
  set suf { " " K M G T}
  return [format {%.3f%s} $free [lrange $suf $indx $indx]]
}
